/*
 *  sydra
 *
 *  sydra is a toolbox which allows you to set up multiple bidirectional
 *  Video/Audio streams from external locations.
 *  sydra has been written to be used for the Elevate Festival in Graz
 *  Austria in order to involve external locations to present themselves
 *  at the festival.
 *
 *
 *  Copyright (C) 2014 Christian Pointner <equinox@spreadspace.org>
 *
 *  This file is part of sydra.
 *
 *  sydra is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  any later version.
 *
 *  sydra is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with sydra. If not, see <http://www.gnu.org/licenses/>.
 *
 *  In addition, as a special exception, the copyright holders hereby
 *  grant permission for non-GPL-compatible GStreamer plugins to be used
 *  and distributed together with GStreamer and sydra.
 *  This permission goes above and beyond the permissions granted by the
 *  GPL license sydra is covered by.
 */

#define _GNU_SOURCE
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>

#include <glib.h>
#include <glib-unix.h>
#include <gst/gst.h>

#include "datatypes.h"
#include "options-rtp.h"
#include "log.h"
#include "daemon.h"
#include "utils.h"
#include "pipelines-rtp.h"
#include "udp.h"


static gboolean sig_handler_terminate(gpointer user_data)
{
  GMainLoop *loop = (GMainLoop *)user_data;

  log_printf(NOTICE, "signal received, closing application");
  g_main_loop_quit(loop);
  return TRUE;
}

static gboolean bus_call(GstBus *bus, GstMessage *msg, gpointer data)
{
  GMainLoop *loop = (GMainLoop *)data;

  switch (GST_MESSAGE_TYPE(msg)) {
  case GST_MESSAGE_EOS: {
    log_printf(NOTICE, "End of stream");
    g_main_loop_quit(loop);
    break;
  }
  case GST_MESSAGE_APPLICATION: {
    log_printf(DEBUG, "Got Application Message!");
    const GstStructure* ms = gst_message_get_structure(msg);
    gboolean quit;
    gst_structure_get_boolean(ms, "quit", &quit);
    if(quit) {
      const gchar* reason = gst_structure_get_string (ms, "reason");
      log_printf(NOTICE, "closing due to message: %s", reason);
      g_main_loop_quit(loop);
    }
    break;
  }
  case GST_MESSAGE_INFO: {
    GError *info;
    gst_message_parse_info(msg, &info, NULL);
    log_printf(INFO, "%s", info->message);
    g_error_free(info);
    break;
  }
  case GST_MESSAGE_WARNING: {
    GError *warning;
    gst_message_parse_warning(msg, &warning, NULL);
    log_printf(WARNING, "%s", warning->message);
    g_error_free(warning);
    break;
  }
  case GST_MESSAGE_ERROR: {
    GError *error;
    gst_message_parse_error(msg, &error, NULL);
    log_printf(ERROR, "%s", error->message);
    g_error_free(error);
    g_main_loop_quit(loop);
    break;
  }
  case GST_MESSAGE_STATE_CHANGED: {
    GstState old_state, new_state;
    gst_message_parse_state_changed(msg, &old_state, &new_state, NULL);
    log_printf(DEBUG, "Element '%s' changed state from %s to %s",
        (msg->src ? GST_OBJECT_NAME(msg->src) : "NULL"),
        gst_element_state_get_name(old_state),
        gst_element_state_get_name(new_state));
    break;
  }
  case GST_MESSAGE_NEW_CLOCK:
  {
    GstClock *clock;
    gst_message_parse_new_clock(msg, &clock);
    log_printf(NOTICE, "New clock: %s", (clock ? GST_OBJECT_NAME (clock) : "NULL"));
    break;
  }
  case GST_MESSAGE_QOS: {
    guint64 running_time, stream_time, timestamp, duration;
    gst_message_parse_qos(msg, NULL, &running_time, &stream_time, &timestamp, &duration);
    log_printf(WARNING, "Element '%s' dropped frames running_time=%lu, stream_time=%lu, timestamp=%lu, duration=%lu",
        (msg->src ? GST_OBJECT_NAME(msg->src) : "NULL"), running_time, stream_time, timestamp, duration);
    break;
  }
  /* case GST_MESSAGE_STREAM_STATUS: */
  /* { */
  /*   GstStreamStatusType type; */
  /*   GstElement *owner; */
  /*   const GValue *val; */
  /*   gchar *path, *ownerstr; */
  /*   GstTask *task = NULL; */

  /*   gst_message_parse_stream_status (msg, &type, &owner); */
  /*   val = gst_message_get_stream_status_object (msg); */

  /*   path = gst_object_get_path_string (GST_MESSAGE_SRC (msg)); */
  /*   ownerstr = gst_object_get_path_string (GST_OBJECT (owner)); */
  /*   log_printf(DEBUG,"Recevied Stream-Status message type: %d, source: %s, owner: %s, object: type %s, value %p", */
  /*                       type, path, ownerstr, G_VALUE_TYPE_NAME (val), g_value_get_object (val)); */
  /*   g_free (path); */
  /*   g_free (ownerstr); */

  /*   /\* see if we know how to deal with this object *\/ */
  /*   if (G_VALUE_TYPE (val) == GST_TYPE_TASK) { */
  /*     task = g_value_get_object (val); */
  /*   } */

  /*   switch (type) { */
  /*     case GST_STREAM_STATUS_TYPE_CREATE: */
  /*       log_printf(DEBUG,"   created task %p", task); */
  /*       break; */
  /*     case GST_STREAM_STATUS_TYPE_ENTER: */
  /*       /\* log_printf(DEBUG,"   raising task priority"); *\/ */
  /*       /\* setpriority (PRIO_PROCESS, 0, -10); *\/ */
  /*       break; */
  /*     case GST_STREAM_STATUS_TYPE_LEAVE: */
  /*       break; */
  /*     default: */
  /*       break; */
  /*   } */
  /*   break; */
  /* } */
  default:
    /* log_printf(DEBUG, "unkonwn message %s from %s", GST_MESSAGE_TYPE_NAME(msg), GST_MESSAGE_SRC_NAME(msg)); */
    return TRUE;
  }
  return TRUE;
}

int main_loop(options_t* opt)
{
  log_printf(NOTICE, "entering main loop");

  struct udp_sinks sinks = { { NULL, NULL }, { NULL, NULL },
                           { NULL, NULL }, { NULL, NULL }, opt->timeout_ };
  struct udp_sources sources = { NULL, NULL, NULL, NULL, opt->keepalive_int_, NULL , opt->rtp_port_base_ };
  GstElement *pipeline = opt->mode_ == SENDER ? create_sender_pipeline(opt, &sinks) :
                                                create_receiver_pipeline(opt, &sources);
  if(!pipeline) {
    log_printf(ERROR, "creating pipeline failed - exitting");
    return -1;
  }

  GMainLoop *loop = g_main_loop_new(NULL, FALSE);

  GstBus *bus = gst_pipeline_get_bus(GST_PIPELINE(pipeline));
  gst_bus_add_watch(bus, bus_call, loop);
  gst_object_unref(GST_OBJECT(bus));

  gulong deep_notify_id = 0;
  if(opt->debug_) {
    deep_notify_id = g_signal_connect(pipeline, "deep-notify",
        G_CALLBACK(gst_object_default_deep_notify), NULL);
  }

  log_printf(NOTICE, "Set State: Paused");
  gst_element_set_state(pipeline, GST_STATE_PAUSED);

  if((opt->mode_ == SENDER && opt->auto_client_ && !attach_udpsinks(&sinks)) ||
     (opt->mode_ == RECEIVER && !attach_udpsources(&sources, opt->rtp_host_)))
    return -1;

  log_printf(NOTICE, "Set State: Playing");
  gst_element_set_state(pipeline, GST_STATE_PLAYING);

  g_unix_signal_add(SIGHUP, sig_handler_terminate, loop);
  g_unix_signal_add(SIGINT, sig_handler_terminate, loop);
  g_unix_signal_add(SIGTERM, sig_handler_terminate, loop);
  g_main_loop_run(loop);

  if(opt->mode_ == RECEIVER)
    detach_udpsources(&sources);

  if (deep_notify_id != 0)
    g_signal_handler_disconnect(pipeline, deep_notify_id);

  log_printf(NOTICE, "Stopping pipeline");
  gst_element_set_state(pipeline, GST_STATE_NULL);
  gst_object_unref(GST_OBJECT(pipeline));

  return 0;
}

int main(int argc, char* argv[])
{
  log_init();
  log_close();

  options_t opt;
  int ret = options_parse(&opt, argc, argv);
  if(ret) {
    if(ret == -1) {
      options_print_version();
      ret = 0;
    }
    options_clear(&opt);
    log_close();
    exit(ret);
  }

  guint len = g_strv_length(opt.log_targets_);
  guint i;
  for(i = 0; i < len; ++i) {
    ret = log_add_target(opt.log_targets_[i]);
    if(ret) {
      switch(ret) {
      case -2: fprintf(stderr, "memory error on log_add_target, exitting\n"); break;
      case -3: fprintf(stderr, "unknown log target: '%s', exitting\n", (char*)(opt.log_targets_[i])); break;
      case -4: fprintf(stderr, "this log target is only allowed once: '%s', exitting\n", (char*)(opt.log_targets_[i])); break;
      default: fprintf(stderr, "syntax error near: '%s', exitting\n", (char*)(opt.log_targets_[i])); break;
      }

      options_clear(&opt);
      log_close();
      exit(ret);
    }
  }

  log_printf(NOTICE, "just started...");

  if(opt.debug_)
    options_print(&opt);

  if(opt.appname_)
    g_set_prgname (opt.appname_);
  else
    g_set_prgname (opt.progname_);

  priv_info_t priv;
  if(opt.username_)
    if(priv_init(&priv, opt.username_, opt.groupname_)) {
      options_clear(&opt);
      log_close();
      exit(-1);
    }

  FILE* pid_file = NULL;
  if(opt.pid_file_) {
    pid_file = fopen(opt.pid_file_, "w");
    if(!pid_file) {
      log_printf(WARNING, "unable to open pid file: %s", strerror(errno));
    }
  }

  if(opt.chroot_dir_)
    if(do_chroot(opt.chroot_dir_)) {
      options_clear(&opt);
      log_close();
      exit(-1);
    }
  if(opt.username_)
    if(priv_drop(&priv)) {
      options_clear(&opt);
      log_close();
      exit(-1);
    }

  if(opt.daemonize_) {
    pid_t oldpid = getpid();
    daemonize();
    log_printf(INFO, "running in background now (old pid: %d)", oldpid);
  }

  if(pid_file) {
    pid_t pid = getpid();
    fprintf(pid_file, "%d", pid);
    fclose(pid_file);
  }

  const gchar *nano_str;
  guint major, minor, micro, nano;
  gst_version(&major, &minor, &micro, &nano);
  if (nano == 1)
    nano_str = " (CVS)";
  else if (nano == 2)
    nano_str = " (Prerelease)";
  else
    nano_str = "";
  log_printf(NOTICE, "sydra-rtp linked against GStreamer %d.%d.%d%s", major, minor, micro, nano_str);

  ret = main_loop(&opt);

  options_clear(&opt);

  log_printf(NOTICE, "sydra-rtp shutdown");

  gst_deinit();
  log_close();

  return ret;
}
